#![allow(dead_code)]
use crate::grpc::challenge::NodeUpdate;
use dashmap::DashMap;
use dashmap::DashSet;
use solana_sdk::signature::keypair::Keypair;
use std::time::SystemTime;

type IP = String;

#[derive(Clone, PartialEq, Debug)]
pub struct NodeInfo {
    pub started_at: SystemTime,
    pub keepalive: SystemTime,
    pub mint_requests: u64,
    pub total_mints: u64,
    pub ratls_conns: u64,
    pub ratls_attacks: u64,
    pub public: bool,
}

/// Keypair must already be known when creating a Store
/// This means the first node of the network creates the key
/// Second node will grab the key from the first node
pub struct Store {
    key: Keypair,
    nodes: DashMap<IP, NodeInfo>,
    conns: DashSet<IP>,
    // TODO: write persistence
    // persistence: FileManager,
}

impl Store {
    pub fn init(key: Keypair) -> Self {
        Self {
            key,
            nodes: DashMap::new(),
            conns: DashSet::new(),
        }
    }

    pub fn get_keypair_base58(&self) -> String {
        self.key.to_base58_string()
    }

    pub async fn add_conn(&self, ip: &str) {
        self.conns.insert(ip.to_string());
    }

    pub async fn delete_conn(&self, ip: &str) {
        self.conns.remove(ip);
    }

    pub async fn get_localhost(&self) -> NodeUpdate {
        // TODO trigger reset_localhost_keys on error instead of expects
        let node = self.nodes.get("localhost").expect("no localhost node");
        NodeUpdate {
            ip: "localhost".to_string(),
            started_at: Some(prost_types::Timestamp::from(node.started_at)),
            keepalive: Some(prost_types::Timestamp::from(SystemTime::now())),
            total_mints: node.total_mints,
            ratls_conns: node.ratls_conns,
            ratls_attacks: node.ratls_attacks,
            public: false,
        }
    }

    /// This returns true if NodeInfo got modified.
    ///
    /// On a side note, there are two types of people in this world:
    ///     1. Those that can extrapolate... WAT?
    pub async fn process_node_update(&self, _node: NodeUpdate) -> bool {
        // TODO: write this function
        true
    }

    pub async fn get_http_node_list(&self) -> Vec<crate::http_server::NodesResp> {
        self.nodes
            .iter()
            .map(|node| crate::http_server::NodesResp {
                ip: node.key().to_string(),
                joined_at: node.value().started_at,
                last_keepalive: node.value().keepalive,
                mints: node.value().total_mints,
                ratls_connections: node.value().ratls_conns,
                ratls_attacks: node.value().ratls_attacks,
                public: node.value().public,
                mint_requests: node.value().mint_requests,
            })
            .collect()
    }

    pub async fn get_grpc_node_list(&self) -> Vec<NodeUpdate> {
        self.nodes
            .iter()
            .map(|node| NodeUpdate {
                ip: node.key().to_string(),
                started_at: Some(prost_types::Timestamp::from(node.value().started_at)),
                keepalive: Some(prost_types::Timestamp::from(node.value().keepalive)),
                total_mints: node.value().total_mints,
                ratls_conns: node.value().ratls_conns,
                ratls_attacks: node.value().ratls_attacks,
                public: node.value().public,
            })
            .collect()
    }

    // returns a random node that does not have an active connection
    pub async fn get_random_node(&self) -> Option<String> {
        use rand::{rngs::OsRng, RngCore};
        let len = self.nodes.len();
        if len == 0 {
            return None;
        }
        let skip = OsRng.next_u64().try_into().unwrap_or(0) % len;
        self.nodes
            .iter()
            .map(|n| n.key().clone())
            .cycle()
            .skip(skip)
            .find(|k| !self.conns.contains(k))
    }
}
